# -*- coding: utf-8 -*-
"""
Created on Fri Feb 21 14:13:29 2020

@author: Prevot
"""
import numpy as np

from guidata.configtools import get_icon
from guidata.qt.QtGui import QAction, QColor, QMenu
from guidata.qt.QtCore import QPoint, Qt, QEvent
from guidata.qthelpers import add_actions, get_std_icon

from guiqwt.config import _
from guiqwt.events import (
    QtDragHandler,
    RectangularSelectionHandler,
    setup_standard_tool_filter,
    EventMatch,
    KeyEventMatch,
)
from guiqwt.signals import (
    SIG_VALIDATE_TOOL,
    SIG_END_RECT,
    SIG_START_TRACKING,
    SIG_MOVE,
    SIG_STOP_NOT_MOVING,
    SIG_STOP_MOVING,
)
from guiqwt.tools import (
    CommandTool,
    DefaultToolbarID,
    RectangularActionTool,
    InteractiveTool,
    ToggleTool,
)
from guiqwt.shapes import EllipseShape, PolygonShape
from guiqwt.baseplot import canvas_to_axes
from grafit.signals import SIG_TRIGGERED, SIG_MODEL_ADDED, SIG_KEY_PRESSED_EVENT

SHAPE_Z_OFFSET = 1000


class KeyMatch(EventMatch):
    """
    A callable returning True if it matches a key event
    keysequence:  integer
    """

    def __init__(self, keysequence):
        super(KeyMatch, self).__init__()
        assert isinstance(keysequence, int)
        self.keyseq = keysequence

    def get_event_types(self):
        return frozenset((QEvent.KeyPress,))

    def __call__(self, event):
        return event.type() == QEvent.KeyPress and event.key() == self.keyseq


class ShowWeightsTool(ToggleTool):
    def __init__(
        self,
        manager,
        title="show weights",
        icon="weight.png",
        tip="show weights associated to the data",
        toolbar_id=DefaultToolbarID,
    ):
        super(ShowWeightsTool, self).__init__(
            manager, title, icon, toolbar_id=toolbar_id
        )

    def activate(self, checked=True):
        self.emit(SIG_VALIDATE_TOOL)


class RunTool(CommandTool):
    def __init__(
        self,
        manager,
        title="Run",
        icon="apply.png",
        tip="run 2D fit",
        toolbar_id=DefaultToolbarID,
    ):
        super(RunTool, self).__init__(manager, title, icon, toolbar_id=toolbar_id)

    def setup_context_menu(self, menu, plot):
        pass

    def activate(self, checked=True):
        self.emit(SIG_VALIDATE_TOOL)


class SaveFitTool(CommandTool):
    def __init__(self, manager, toolbar_id=DefaultToolbarID):
        super(SaveFitTool, self).__init__(
            manager,
            _("Save fit result"),
            get_std_icon("DialogSaveButton", 16),
            toolbar_id=toolbar_id,
        )

    def activate(self, checked=True):
        self.emit(SIG_VALIDATE_TOOL)


class PrefTool(CommandTool):
    def __init__(self, manager, toolbar_id=DefaultToolbarID):
        CommandTool.__init__(
            self,
            manager,
            _("Run"),
            icon="settings.png",
            tip=_("Preferences"),
            toolbar_id=toolbar_id,
        )
        self.manager = manager

    def create_action_menu(self, manager):
        # Create and return menu for the tool's action
        self.saveprefaction = QAction("Save as...", self)
        self.showtagsaction = QAction("Show tags", self)
        self.scaleprefaction = QAction("Autoscale", self)
        self.scaleprefaction.setCheckable(True)
        self.scaleprefaction.setChecked(True)
        self.restartaction = QAction("Restart from last parameters saved", self)
        self.restartaction.setCheckable(True)
        self.restartaction.setChecked(True)
        self.samescaleaction = QAction("Same scale for data and model images", self)
        self.samescaleaction.setCheckable(True)
        self.samescaleaction.setChecked(True)
        self.showdiffaction = QAction("Show difference profiles", self)
        self.showdiffaction.setCheckable(True)
        self.showdiffaction.setChecked(True)

        menu = QMenu()
        add_actions(
            menu,
            (
                self.saveprefaction,
                self.showtagsaction,
                self.scaleprefaction,
                self.restartaction,
                self.samescaleaction,
                self.showdiffaction,
            ),
        )

        self.action.setMenu(menu)
        return menu


class EllipseSelectionHandlerCXY(RectangularSelectionHandler):
    # on utilise la classe heritee dont on surcharge la methode move

    def __init__(self, filter, btn, mods=Qt.NoModifier, start_state=0):
        super(EllipseSelectionHandlerCXY, self).__init__(
            filter=filter, btn=btn, mods=mods, start_state=start_state
        )
        filter.add_event(
            start_state, KeyEventMatch((42, 43, 45, 47)), self.key_press, start_state
        )

    def move(self, filter, event):
        """methode surchargee par la classe """
        x1, y1 = canvas_to_axes(self.shape, event.pos())
        x0, y0 = canvas_to_axes(self.shape, self.start)
        dx = x1 - x0
        dy = y1 - y0
        self.shape.points = np.array(
            [[x0 - dx, y0], [x0 + dx, y0], [x0, y0 - dy], [x0, y0 + dy]]
        )
        self.move_action(filter, event)
        filter.plot.replot()

    def key_press(self, filter, event):
        self.emit(SIG_KEY_PRESSED_EVENT, filter, event.key())

    def set_shape(
        self, shape, h0, h1, h2, h3, setup_shape_cb=None, avoid_null_shape=False
    ):
        self.shape = shape
        self.shape_h0 = h0
        self.shape_h1 = h1
        self.shape_h2 = h2
        self.shape_h3 = h3
        self.setup_shape_cb = setup_shape_cb
        self.avoid_null_shape = avoid_null_shape


class RectangularSelectionHandlerCXY(RectangularSelectionHandler):
    # on utilise la classe heritee dont on surcharge la methode move

    def move(self, filter, event):
        """methode surchargee par la classe """
        sympos = QPoint(
            2 * self.start.x() - event.pos().x(), 2 * self.start.y() - event.pos().y()
        )
        print(self.shape_h0, self.shape_h1)
        self.shape.move_local_point_to(self.shape_h0, sympos)
        self.shape.move_local_point_to(self.shape_h1, event.pos())
        self.move_action(filter, event)
        filter.plot.replot()


class RectangularActionToolCXY(RectangularActionTool):
    # outil de tracage de rectangles centre en x
    # on utilise la classe heritee dont on surcharge la methode setup_filter

    def setup_filter(
        self, baseplot
    ):  # partie utilisee pendant le mouvement a la souris
        # utilise a l'initialisation de la toolbar
        # print "setup_filter"
        filter = baseplot.filter
        start_state = filter.new_state()
        handler = RectangularSelectionHandlerCXY(
            filter, Qt.LeftButton, start_state=start_state  # gestionnaire du filtre
        )
        shape, h0, h1 = self.get_shape()
        shape.pen.setColor(QColor("#00bfff"))

        handler.set_shape(
            shape, h0, h1, self.setup_shape, avoid_null_shape=self.AVOID_NULL_SHAPE
        )
        self.connect(handler, SIG_END_RECT, self.end_rect)
        # self.connect(handler, SIG_CLICK_EVENT, self.start)   #a definir aussi dans RectangularSelectionHandler2
        return setup_standard_tool_filter(filter, start_state)

    def activate(self):
        """Activate tool"""

        # print "commande active",self
        for baseplot, start_state in self.start_state.items():
            baseplot.filter.set_state(start_state, None)
        self.action.setChecked(True)
        self.manager.set_active_tool(self)
        # plot = self.get_active_plot()
        # plot.newcurve=True

    def deactivate(self):
        """Deactivate tool"""
        # print "commande desactivee",self
        self.action.setChecked(False)


class CircularActionToolCXY(RectangularActionToolCXY):
    TITLE = _("Circle")
    ICON = "circle.png"

    def __init__(
        self,
        manager,
        func1,
        func2,
        shape_style=None,
        toolbar_id=DefaultToolbarID,
        title=None,
        icon=None,
        tip=None,
        fix_orientation=False,
        switch_to_default_tool=None,
    ):

        self.key_func = func2  # function for keys
        super(CircularActionToolCXY, self).__init__(
            manager,
            func1,
            shape_style=shape_style,
            toolbar_id=toolbar_id,
            title=title,
            icon=icon,
            tip=tip,
            fix_orientation=fix_orientation,
            switch_to_default_tool=switch_to_default_tool,
        )

    def setup_filter(
        self, baseplot
    ):  # partie utilisee pendant le mouvement a la souris
        # utilise a l'initialisation de la toolbar
        # print "setup_filter"
        filter = baseplot.filter
        start_state = filter.new_state()
        handler = EllipseSelectionHandlerCXY(
            filter, Qt.LeftButton, start_state=start_state  # gestionnaire du filtre
        )
        shape, h0, h1, h2, h3 = self.get_shape()
        shape.pen.setColor(QColor("#00bfff"))

        handler.set_shape(
            shape,
            h0,
            h1,
            h2,
            h3,
            self.setup_shape,
            avoid_null_shape=self.AVOID_NULL_SHAPE,
        )
        self.connect(handler, SIG_END_RECT, self.end_rect)
        self.connect(handler, SIG_KEY_PRESSED_EVENT, self.key_pressed)
        return setup_standard_tool_filter(filter, start_state)

    def key_pressed(self, filter, key):
        plot = filter.plot
        self.key_func(plot, key)

    def get_shape(self):
        """Reimplemented RectangularActionTool method"""
        shape, h0, h1, h2, h3 = self.create_shape()
        self.setup_shape(shape)
        return shape, h0, h1, h2, h3

    def create_shape(self):
        shape = EllipseShape(0, 0, 1, 1)
        self.set_shape_style(shape)
        shape.switch_to_ellipse()
        return shape, 0, 1, 2, 3  # give the values of h0,h1,h2,h3 use to move the shape


class MultiPointTool(InteractiveTool):
    TITLE = _("Polyline")
    ICON = "polyline.png"
    CURSOR = Qt.ArrowCursor

    def __init__(
        self,
        manager,
        func,
        handle_final_shape_cb=None,
        shape_style=None,
        toolbar_id=DefaultToolbarID,
        title=None,
        icon=None,
        tip=None,
        switch_to_default_tool=None,
    ):
        super(MultiPointTool, self).__init__(
            manager,
            toolbar_id,
            title=title,
            icon=icon,
            tip=tip,
            switch_to_default_tool=switch_to_default_tool,
        )
        self.handle_final_shape_cb = handle_final_shape_cb
        self.shape = None
        self.current_handle = None
        self.init_pos = None
        self.move_func = func
        if shape_style is not None:
            self.shape_style_sect = shape_style[0]
            self.shape_style_key = shape_style[1]
        else:
            self.shape_style_sect = "plot"
            self.shape_style_key = "shape/drag"

    def reset(self):
        self.shape = None
        self.current_handle = None

    def create_shape(self, filter, pt):
        self.shape = PolygonShape(closed=False)
        filter.plot.add_item_with_z_offset(self.shape, SHAPE_Z_OFFSET)
        self.shape.setVisible(True)
        self.shape.set_style(self.shape_style_sect, self.shape_style_key)
        self.shape.add_local_point(pt)
        return self.shape.add_local_point(pt)

    def setup_filter(self, baseplot):
        filter = baseplot.filter
        # Initialisation du filtre
        start_state = filter.new_state()
        # Bouton gauche :
        handler = QtDragHandler(filter, Qt.LeftButton, start_state=start_state)
        filter.add_event(
            start_state,
            KeyEventMatch((Qt.Key_Enter, Qt.Key_Return, Qt.Key_Space)),
            self.validate,
            start_state,
        )
        filter.add_event(
            start_state,
            KeyEventMatch((Qt.Key_Backspace, Qt.Key_Escape,)),
            self.cancel_point,
            start_state,
        )
        self.connect(handler, SIG_START_TRACKING, self.mouse_press)
        self.connect(handler, SIG_MOVE, self.move)
        self.connect(handler, SIG_STOP_NOT_MOVING, self.mouse_release)
        self.connect(handler, SIG_STOP_MOVING, self.mouse_release)
        return setup_standard_tool_filter(filter, start_state)

    def validate(self, filter, event):
        super(MultiPointTool, self).validate(filter, event)
        """
        if self.handle_final_shape_cb is not None:
            self.handle_final_shape_cb(self.shape)
        """
        self.shape.detach()
        filter.plot.replot()
        self.reset()

    def cancel_point(self, filter, event):
        if self.shape is None:
            return
        points = self.shape.get_points()
        if points is None:
            return
        elif len(points) <= 2:
            filter.plot.del_item(self.shape)
            self.reset()
        else:
            if self.current_handle:
                newh = self.shape.del_point(self.current_handle)
            else:
                newh = self.shape.del_point(-1)
            self.current_handle = newh
        filter.plot.replot()

    def mouse_press(self, filter, event):
        """We create a new shape if it's the first point
        otherwise we add a new point
        """
        if self.shape is None:
            self.init_pos = event.pos()
            self.current_handle = self.create_shape(filter, event.pos())
            filter.plot.replot()
        else:
            self.current_handle = self.shape.add_local_point(event.pos())
            if self.current_handle > 2:
                self.shape.del_point(0)
                self.current_handle = 2

    def move(self, filter, event):
        """moving while holding the button down lets the user
        position the last created point
        """
        if self.shape is None or self.current_handle is None:
            # Error ??
            return
        self.shape.move_local_point_to(self.current_handle, event.pos())
        self.move_func(filter.plot, self.shape.points)
        filter.plot.replot()

    def mouse_release(self, filter, event):
        """Releasing the mouse button validate the last point position"""
        if self.current_handle is None:
            return
        if self.init_pos is not None and self.init_pos == event.pos():
            self.shape.del_point(-1)
        else:
            self.shape.move_local_point_to(self.current_handle, event.pos())
        self.init_pos = None
        self.current_handle = None
        self.move_func(filter.plot, self.shape.points)
        filter.plot.replot()

    def deactivate(self):
        """Deactivate tool"""
        if self.shape is not None:
            self.shape.detach()
            self.get_active_plot().replot()
            self.reset()
        self.action.setChecked(False)


class AddModelTool(CommandTool):
    def __init__(self, manager, list_of_models, toolbar_id=DefaultToolbarID):
        self.list_of_models = list_of_models
        CommandTool.__init__(
            self,
            manager,
            _("Fit"),
            icon=get_icon("edit_add.png"),
            tip=_("Add curve for fit"),
            toolbar_id=toolbar_id,
        )
        self.manager = manager

    def create_action_menu(self, manager):
        # Create and return menu for the tool's action
        menu = QMenu()
        for i, modelclass in enumerate(self.list_of_models):
            action = QAction(get_icon(modelclass.ICON), modelclass.NAME, self)
            add_actions(menu, (action,))
            self.connect(action, SIG_TRIGGERED, lambda arg=i: self.add_model(arg))

        self.action.setMenu(menu)
        return menu

    def add_model(self, k):
        print(k)
        i = k
        self.emit(SIG_MODEL_ADDED, i)



"""
 This file is part of the binoculars project.

  The BINoculars library is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  The BINoculars library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with the hkl library.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2012-2015 European Synchrotron Radiation Facility
                          Grenoble, France

  Authors: Willem Onderwaater <onderwaa@esrf.fr>
           Jonathan Rawle

"""
import sys
import os
import itertools
import numpy
import time
import math
import json

from functools import reduce

from scipy.misc import imread
import scisoftpy as dnp

from scisoftpy import sin, cos

from .. import backend, errors, util


class HKLProjection(backend.ProjectionBase):
    # scalars: mu, theta, [chi, phi, "omitted"] delta, gamR, gamT, ty, wavelength
    # 3x3 matrix: UB
    def project(self, energy, UB, pixels, gamma, delta, omega, alpha, nu):
        # put the detector at the right position

        dx, dy, dz = pixels

        # convert angles to radians
        gamma, delta, alpha, omega, nu = numpy.radians((gamma, delta, alpha, omega, nu))

        RGam = numpy.matrix(
            [[1, 0, 0], [0, cos(gamma), -sin(gamma)], [0, sin(gamma), cos(gamma)]]
        )
        RDel = (
            numpy.matrix(
                [[cos(delta), -sin(delta), 0], [sin(delta), cos(delta), 0], [0, 0, 1]]
            )
        ).getI()
        RNu = numpy.matrix([[cos(nu), 0, sin(nu)], [0, 1, 0], [-sin(nu), 0, cos(nu)]])

        # calculate Cartesian coordinates for each pixel using clever matrix stuff
        M = numpy.mat(
            numpy.concatenate((dx.flatten(0), dy.flatten(0), dz.flatten(0))).reshape(
                3, dx.shape[0] * dx.shape[1]
            )
        )
        XYZp = RGam * RDel * RNu * M
        xp = dnp.array(XYZp[0]).reshape(dx.shape)
        yp = dnp.array(XYZp[1]).reshape(dy.shape)
        zp = dnp.array(XYZp[2]).reshape(dz.shape)
        # don't bother with the part about slits...

        # Calculate effective gamma and delta for each pixel
        d_ds = dnp.sqrt(xp ** 2 + yp ** 2 + zp ** 2)
        Gam = dnp.arctan2(zp, yp)
        Del = -1 * dnp.arcsin(-xp / d_ds)

        # wavenumber
        k = 2 * math.pi / 12.398 * energy

        # Define the needed matrices. The notation follows the article by Bunk &
        # Nielsen. J.Appl.Cryst. (2004) 37, 216-222.
        M1 = k * numpy.matrix(
            cos(omega) * sin(Del)
            - sin(omega)
            * (
                cos(alpha) * (cos(Gam) * cos(Del) - 1)
                + sin(alpha) * sin(Gam) * cos(Del)
            )
        )
        M2 = k * numpy.matrix(
            sin(omega) * sin(Del)
            + cos(omega)
            * (
                cos(alpha) * (cos(Gam) * cos(Del) - 1)
                + sin(alpha) * sin(Gam) * cos(Del)
            )
        )
        M3 = k * numpy.matrix(
            -sin(alpha) * (cos(Gam) * cos(Del) - 1) + cos(alpha) * sin(Gam) * cos(Del)
        )

        # invert UB matrix
        UBi = numpy.matrix(UB).getI()

        # calculate HKL
        H = UBi[0, 0] * M1 + UBi[0, 1] * M2 + UBi[0, 2] * M3
        K = UBi[1, 0] * M1 + UBi[1, 1] * M2 + UBi[1, 2] * M3
        L = UBi[2, 0] * M1 + UBi[2, 1] * M2 + UBi[2, 2] * M3

        return (H, K, L)

    def get_axis_labels(self):
        return "H", "K", "L"


class GammaDelta(
    HKLProjection
):  # just passing on the coordinates, makes it easy to accurately test the theta correction
    def project(self, beamenergy, UB, gamma, delta, omega, alpha):
        delta, gamma = numpy.meshgrid(delta, gamma)
        return (gamma, delta)

    def get_axis_labels(self):
        return "Gamma", "Delta"


class pixels(backend.ProjectionBase):
    def project(self, beamenergy, UB, gamma, delta, omega, alpha):
        y, x = numpy.mgrid[slice(None, gamma.shape[0]), slice(None, delta.shape[0])]
        return (y, x)

    def get_axis_labels(self):
        return "y", "x"


class IO7Input(backend.InputBase):
    # OFFICIAL API

    dbg_scanno = None
    dbg_pointno = None

    def generate_jobs(self, command):
        scans = util.parse_multi_range(",".join(command).replace(" ", ","))
        if not len(scans):
            sys.stderr.write("error: no scans selected, nothing to do\n")
        for scanno in scans:
            util.status("processing scan {0}...".format(scanno))
            if self.config.pr:
                pointcount = self.config.pr[1] - self.config.pr[0] + 1
                start = self.config.pr[0]
            else:
                scan = self.get_scan(scanno)
                pointcount = len(scan.file)
                start = 0
            if pointcount > self.config.target_weight * 1.4:
                for s in util.chunk_slicer(pointcount, self.config.target_weight):
                    yield backend.Job(
                        scan=scanno,
                        firstpoint=start + s.start,
                        lastpoint=start + s.stop - 1,
                        weight=s.stop - s.start,
                    )
            else:
                yield backend.Job(
                    scan=scanno,
                    firstpoint=start,
                    lastpoint=start + pointcount - 1,
                    weight=pointcount,
                )

    def process_job(self, job):
        super(IO7Input, self).process_job(job)
        scan = self.get_scan(job.scan)
        self.metadict = dict()
        try:
            scanparams = self.get_scan_params(scan)  # wavelength, UB
            pointparams = self.get_point_params(
                scan, job.firstpoint, job.lastpoint
            )  # 2D array of diffractometer angles + mon + transm
            images = self.get_images(scan, job.firstpoint, job.lastpoint)  # iterator!
            for pp, image in zip(pointparams, images):
                yield self.process_image(scan, scanparams, pp, image)
            util.statuseol()
        except Exception as exc:
            exc.args = errors.addmessage(
                exc.args,
                ", An error occured for scan {0} at point {1}. See above for more information".format(
                    self.dbg_scanno, self.dbg_pointno
                ),
            )
            raise
        self.metadata.add_section("id7_backend", self.metadict)

    def get_scan_params(self, scan):
        energy = scan.metadata.dcm1energy
        UB = numpy.array(json.loads(scan.metadata.diffcalc_ub))

        self.metadict["UB"] = UB
        self.metadict["energy"] = energy

        return energy, UB

    def get_point_params(self, scan, first, last):
        sl = slice(first, last + 1)

        GAM, DEL, OMG, CHI, PHI, ALF, MON, TRANSM = list(range(8))
        params = numpy.zeros(
            (last - first + 1, 8)
        )  # gamma delta theta chi phi mu mon transm
        params[:, CHI] = 0
        params[:, PHI] = 0

        params[:, OMG] = scan["omega"][sl]
        params[:, GAM] = scan["gamma"][sl]
        params[:, DEL] = scan["delta"][sl]
        params[:, ALF] = scan["alpha"][sl]

        return params

    def get_images(self, scan, first, last, dry_run=False):
        sl = slice(first, last + 1)
        for fn in scan.file[sl]:
            yield imread(self.get_imagefilename(fn))

    def get_imagefilename(self, filename):
        if self.config.imagefolder is None:
            if os.path.exists(filename):
                return filename
            else:
                raise errors.ConfigError(
                    "image filename specified in the datafile does not exist '{0}'".format(
                        filename
                    )
                )
        else:
            head, tail = os.path.split(filename)
            folders = head.split("/")
            try:
                imagefolder = self.config.imagefolder.format(
                    folders=folders, rfolders=list(reversed(folders))
                )
            except Exception as e:
                raise errors.ConfigError(
                    "invalid 'imagefolder' specification '{0}': {1}".format(
                        self.config.imagefolder, e
                    )
                )
            else:
                if not os.path.exists(imagefolder):
                    raise errors.ConfigError(
                        "invalid 'imagefolder' specification '{0}'. Path {1} does not exist".format(
                            self.config.imagefolder, imagefolder
                        )
                    )
            fn = os.path.join(imagefolder, tail)
            if os.path.exists(fn):
                return fn
            else:
                raise errors.ConfigError(
                    "image filename does not exist '{0}', either imagefolder is wrongly specified or image file does not exist".format(
                        filename
                    )
                )

    def parse_config(self, config):
        super(IO7Input, self).parse_config(config)
        self.config.xmask = util.parse_multi_range(
            config.pop("xmask", None)
        )  # Optional, select a subset of the image range in the x direction. all by default
        self.config.ymask = util.parse_multi_range(
            config.pop("ymask", None)
        )  # Optional, select a subset of the image range in the y direction. all by default
        self.config.datafilefolder = config.pop(
            "datafilefolder"
        )  # Folder with the datafiles
        self.config.imagefolder = config.pop(
            "imagefolder", None
        )  # Optional, takes datafile folder tag by default
        self.config.pr = config.pop("pr", None)  # Optional, all range by default
        if self.config.xmask is None:
            self.config.xmask = slice(None)
        if self.config.ymask is None:
            self.config.ymask = slice(None)
        if self.config.pr:
            self.config.pr = util.parse_tuple(self.config.pr, length=2, type=int)
        self.config.centralpixel = util.parse_tuple(
            config.pop("centralpixel"), length=2, type=int
        )  # x,y
        self.config.maskmatrix = config.pop(
            "maskmatrix", None
        )  # Optional, if supplied pixels where the mask is 0 will be removed
        self.config.pixelsize = util.parse_tuple(
            config.pop("pixelsize"), length=2, type=float
        )  # pixel size x/y (mm) (same dimension as sdd)

    def get_destination_options(self, command):
        if not command:
            return False
        command = ",".join(command).replace(" ", ",")
        scans = util.parse_multi_range(command)
        return dict(
            first=min(scans),
            last=max(scans),
            range=",".join(str(scan) for scan in scans),
        )

    # CONVENIENCE FUNCTIONS
    def get_scan(self, scanno):
        filename = os.path.join(self.config.datafilefolder, str(scanno) + ".dat")
        if not os.path.exists(filename):
            raise errors.ConfigError(
                "datafile filename does not exist: {0}".format(filename)
            )
        return dnp.io.load(filename)

    @staticmethod
    def apply_mask(data, xmask, ymask):
        roi = data[ymask, :]
        return roi[:, xmask]


class EH2(IO7Input):
    def parse_config(self, config):
        super(IO7Input, self).parse_config(config)
        self.config.sdd = float(
            config.pop("sdd"), None
        )  # Sample to detector distance (mm)
        if self.config.sdd is not None:
            self.config.sdd = float(self.config.sdd)

    def process_image(self, scan, scanparams, pointparams, image):
        (
            gamma,
            delta,
            omega,
            chi,
            phi,
            alpha,
            mon,
            transm,
        ) = pointparams  # GAM, DEL, OMG, CHI, PHI, ALF, MON, TRANSM
        energy, UB = scanparams

        weights = numpy.ones_like(image)

        util.status(
            "{4}| gamma: {0}, delta: {1}, omega: {2}, mu: {3}".format(
                gamma, delta, omega, alpha, time.ctime(time.time())
            )
        )

        # pixels to angles
        pixelsize = numpy.array(self.config.pixelsize)

        if self.config.sdd is None:
            sdd = scan.metadata.diff1detdist
        else:
            sdd = self.config.sdd

        nu = scan.metadata.diff2prot

        centralpixel = self.config.centralpixel  # (column, row) = (delta, gamma)

        dz = (numpy.indices(image.shape)[1] - centralpixel[1]) * pixelsize[1]
        dx = (numpy.indices(image.shape)[0] - centralpixel[0]) * pixelsize[0]
        dy = numpy.ones(image.shape) * sdd

        # masking
        if self.config.maskmatrix is not None:
            if self.config.maskmatrix.shape != data.shape:
                raise errors.BackendError(
                    "The mask matrix does not have the same shape as the images"
                )
            weights *= self.config.maskmatrix

        intensity = self.apply_mask(image, self.config.xmask, self.config.ymask)
        weights = self.apply_mask(weights, self.config.xmask, self.config.ymask)
        dx = self.apply_mask(dx, self.config.xmask, self.config.ymask)
        dy = self.apply_mask(dy, self.config.xmask, self.config.ymask)
        dz = self.apply_mask(dz, self.config.xmask, self.config.ymask)

        # X,Y = numpy.meshgrid(x,y)
        # Z = numpy.ones(X.shape) * sdd

        pixels = dx, dy, dz

        return intensity, weights, (energy, UB, pixels, gamma, delta, omega, alpha, nu)


class EH1(IO7Input):
    def parse_config(self, config):
        super(EH1, self).parse_config(config)
        self.config.sdd = float(config.pop("sdd"))  # Sample to detector distance (mm)

    def process_image(self, scan, scanparams, pointparams, image):
        (
            gamma,
            delta,
            omega,
            chi,
            phi,
            alpha,
            mon,
            transm,
        ) = pointparams  # GAM, DEL, OMG, CHI, PHI, ALF, MON, TRANSM
        energy, UB = scanparams

        weights = numpy.ones_like(image)

        util.status(
            "{4}| gamma: {0}, delta: {1}, omega: {2}, mu: {3}".format(
                gamma, delta, omega, alpha, time.ctime(time.time())
            )
        )

        # pixels to angles
        pixelsize = numpy.array(self.config.pixelsize)

        sdd = self.config.sdd

        nu = scan.metadata.diff1prot

        centralpixel = self.config.centralpixel  # (column, row) = (delta, gamma)

        dz = (numpy.indices(image.shape)[1] - centralpixel[1]) * pixelsize[1]
        dx = (numpy.indices(image.shape)[0] - centralpixel[0]) * pixelsize[0]
        dy = numpy.ones(image.shape) * sdd

        # masking
        if self.config.maskmatrix is not None:
            if self.config.maskmatrix.shape != data.shape:
                raise errors.BackendError(
                    "The mask matrix does not have the same shape as the images"
                )
            weights *= self.config.maskmatrix

        intensity = self.apply_mask(image, self.config.xmask, self.config.ymask)
        weights = self.apply_mask(weights, self.config.xmask, self.config.ymask)
        dx = self.apply_mask(dx, self.config.xmask, self.config.ymask)
        dy = self.apply_mask(dy, self.config.xmask, self.config.ymask)
        dz = self.apply_mask(dz, self.config.xmask, self.config.ymask)

        pixels = dx, dy, dz

        return intensity, weights, (energy, UB, pixels, gamma, delta, omega, alpha, nu)


def load_matrix(filename):
    if filename == None:
        return None
    if os.path.exists(filename):
        ext = os.path.splitext(filename)[-1]
        if ext == ".txt":
            return numpy.array(numpy.loadtxt(filename), dtype=numpy.bool)
        elif ext == ".npy":
            return numpy.array(numpy.load(filename), dtype=numpy.bool)
        else:
            raise ValueError(
                "unknown extension {0}, unable to load matrix!\n".format(ext)
            )
    else:
        raise IOError(
            "filename: {0} does not exist. Can not load matrix".format(filename)
        )
